from collections import OrderedDict
from datetime import datetime
from duckietown_swarm.constants import DSwarmConstants
from duckietown_swarm.utils import as_seconds, now_until
import os
import shutil
from tempfile import mkdtemp
import time

from conf_tools.utils import locate_files
from system_cmd import CmdException

from .dcache_wire import create_propose_message
from .ipfs_utils import IPFSInterface
from .utils import yaml_dump_omaps


def directory_watcher(mi, watch_dir):
    ipfsi = IPFSInterface(mi.ipfs_path)
    seen = {}
    tmpdir = mkdtemp(prefix='swarm')

    def look_for_files(d, pattern='*'):
        if not os.path.exists(d):
            msg = 'Logs directory does not exist: %s' % d
            raise ValueError(msg)

        filenames = locate_files(d, pattern)
        for f in filenames:
            if 'cache' in f:
                continue
            if f.startswith('.'):
                continue

            if not f in seen:
                seen[f] = True  # XXX

                try:
                    hashed = _add_file(f)
                except Exception as e:
                    print('error adding %s:\n%s' % (f, e))
                    continue

                print('found local %s %s' % (hashed, f))

                interval = now_until(as_seconds('7d'))
                msg = create_propose_message(DSwarmConstants.DLC_BUCKET_FILES, hashed, interval)
                mi.send_to_brain('', msg, sign=True)

    def _add_file(filename):
        d = setup_dir_for_file(ipfsi, tmpdir, filename)
        try:
            hashed = ipfsi.add_ipfs_dir(d)
            return hashed
        except CmdException as e:
            raise Exception(str(e))

    while True:
        look_for_files(watch_dir)
        time.sleep(59)


def setup_dir_for_file(ipfsi, tmpdir, filename):
    b0 = os.path.basename(filename)
    b = b0.replace('.', '_')
    d = os.path.join(tmpdir, b)
    if not os.path.exists(d):
        os.mkdir(d)
    dest = os.path.join(d, b0)

    if not os.path.exists(dest):
        try:
            os.link(filename, dest)
        except OSError:
            shutil.copy(filename, dest)

    r = OrderedDict()
    r['filename'] = b0
    mtime = os.path.getmtime(filename)
    r['ctime'] = datetime.fromtimestamp(mtime)
    # no - otherwise it changes every time
#    r['upload_host'] = socket.gethostname()
    # r['upload_date'] = datetime.now()
    r['upload_node'] = ipfsi.ipfs_id()
#    r['upload_user'] = getpass.getuser()
    info = os.path.join(d, 'upload_info1.yaml')
    with open(info, 'w') as f:
        y = yaml_dump_omaps(r)
        f.write(y)
    return d




#! /usr/bin/env python

from collections import defaultdict
from duckietown_swarm.dcache_wire import create_ping_message
import random
import socket
from tempfile import mkdtemp
import time

from contracts import contract
from contracts.utils import indent
from irc.bot import SingleServerIRCBot
from irc.client import ServerNotConnectedError, MessageTooLong

from . import __version__
from .dcache import  create_propose_message
from .dcache import Envelope, CouldNotReadEnvelope
from .fragments import fragmentize, Assembler
from .ipfs_utils import IPFSInterface


class SwarmBot(SingleServerIRCBot):

    def __init__(self, servers, stream_name, mi, channel, nickname):
        SingleServerIRCBot.__init__(self, servers, nickname, nickname)
        self.mi = mi
        self.target = channel

        self.stream_name = stream_name
        self.seen = {}
        self.tmpdir = mkdtemp(prefix='swarm')

        self.admitted = False
        self.disconnected = False
        self.last_broadcast = 0

        self.lateness = random.uniform(0, 10)
        self.nusers = 10

        self.nreceived = 0
        self.assemblers = defaultdict(Assembler)

        ipfsi = mi.get_ipfsi()
        self.identity = ipfsi.ipfs_id()

    def on_pubmsg(self, c, e):
        self._handle(c, e)

    def on_privmsg(self, c, e):
        self._handle(c, e)

    def on_namreply(self, _c, e):
        names = map(str, e.arguments[2].split())
        print('list of handles: %s' % names)
        for name in names:
            e = Envelope(
                    maddr_from='/dswarm/%s/brain' % self.identity,
                    maddr_to='/dswarm/*/brain',
                    contents=create_ping_message())
            send(self, e, name)

    def _handle(self, _c, e):
        self.nreceived += 1
        s = str(e.arguments[0])
        # {'source': u'andrea!~andrea@nutonomy02.n.subnet.rcn.com',
        # 'tags': [], 'type': 'pubmsg', 'target': u'#duckiebots', 'arguments': [u'deplep']}
        source = e.source
        if '!' in source:
            source = source[:source.index('!')]

        sn = '/' + str(source)

        self.assemblers[sn].push(s)
        msgs = self.assemblers[sn].pop()
        for msg in msgs:
            try:
                envelope = Envelope.from_json(msg)
            except CouldNotReadEnvelope as e:
                em = 'Could not read envelope from %s: %s' % (sn, e)
                print(em)
            else:
                envelope.maddr_from = sn + envelope.maddr_from
                from duckietown_swarm.irc2 import CouldNotDispatch

                try:
                    self.mi.dispatch(envelope)
                except CouldNotDispatch as e:
                    print(s)

    def on_welcome(self, connection, _event):
#        print('welcome message from %s: \n%s' % (connection.server, " ".join(event.arguments)))
        connection.join(self.target)
        connection.set_keepalive(30)
        print('Connected to %s on IRC server %s.' % (self.target, connection.server))
        self.admitted = True

    def on_join(self, c, e):
        nick = e.source.nick
#        print('on_join %s' % nick)
        if nick != c.get_nickname():
            e = Envelope(
                    maddr_from='/dswarm/%s/brain' % self.identity,
                    maddr_to='/dswarm/*/brain',
                    contents=create_ping_message())
            send(self, e, nick)
#            self.channels[ch] = Channel()
#
#        users = self.channels[self.target].users()
#        self.nusers = len(users)
#
#        if False:
#            msg = create_request_message(BOOTSTRAP_PATTERNS, validity=now_until(105))
#            maddr_from = '/dswarm/%s/brain' % self.identity
#            envelope = Envelope(maddr_from=maddr_from,
#                         maddr_to='/dswarm/*/brain',
#                         contents=msg)
#            send(self, envelope, self.target)
##        print('current users: %s' % users)

    def on_disconnect(self, _c, event):
        print('Disconnected', event.__dict__)
        self.disconnected = True

    @contract(d=str)
    def send_message(self, target, d):
        self.connection.privmsg(target, d)
#        print('to %s: %s' % (target, d))
        self.last_broadcast = time.time()


def start_irc(mi, server, channel):
    random.seed()

    ipfsi = IPFSInterface(mi.ipfs_path)
    identity = ipfsi.ipfs_id()
    hostname = socket.gethostname()
    printable = ''.join(ch for ch in hostname if ch.isalnum())
    printable = printable[:8]
    nickname = '%s_' % (__version__[-3:].replace('.', '')) + printable + '_' + identity[-4:]
    print('Using nickname %s' % nickname)

    delta_wait_for_welcome = 60
    stream_name = '/dns4/' + server[0] + '/tcp/' + str(server[1]) + '/irc'
    assert stream_name in mi.name2queue
    assert stream_name == mi.my_maddr

#    key0 = CryptoKey(str(random.randint(0, 10000)), str(random.randint(0, 10000)))
#    key0.hash = ipfsi.ipfs_id()

    t = time.time()
    irc_names = [
        stream_name + '/' + nickname,
        stream_name + '/' + channel,
    ]
    for irc_name in irc_names:
        msg = create_propose_message(('pri', identity, 'addresses'), irc_name, validity=[t, t + 300])
        mi.send_to_brain('', msg, sign=True)

    c = SwarmBot([server], stream_name, mi, channel, nickname)
    while True:

        while True:
            attempt_start = time.time()
            c.admitted = False
            print('server list: %s' % [_.host for _ in c.server_list])
            c._connect()
            while not c.admitted:
                # print('Waiting for welcome')
                c.reactor.process_once(1.0)

                delta = time.time() - attempt_start
                give_up = delta > delta_wait_for_welcome
                if give_up:
                    print('Could not receive welcome. giving up after %d s'
                          % delta_wait_for_welcome)
                    break
            if c.admitted:
                break
            print('Changing server')
            c.jump_server()

        c.disconnected = False

        while True:
#            print('processing once')
            c.reactor.process_once()
            if c.disconnected:
                break

            xs = mi.get_many_for_me(timeout=1.0)
#            print('get_many_for_me %s' % len(xs))
            for envelope in xs:
                try:
                    if not envelope.maddr_to or envelope.maddr_to[0] != '/':
                        msg = 'Invalid addr_to: %r' % envelope.maddr_to
                        msg += '\n\n' + indent(envelope, '  ')
                        print(msg)
                        continue

                    # orig = envelope.maddr_to
                    tokens = envelope.maddr_to.split('/')
                    assert tokens[0] == ''  # starts with '/'
                    dest = tokens[1]  # = '#duckiebots'
                    rest = '/' + "/".join(tokens[2:])
                    envelope.maddr_to = rest

#                    print('orig: %s  new: %s' % (orig, envelope.maddr_to))
                    if False:
                        print('not sending: %s' % envelope.contents)
                    else:
                        send(c, envelope, dest)

#                    mi.send_to_brain(envelope.maddr_to, envelope.contents)
                except ServerNotConnectedError as e:
                    print e
                    break


@contract(envelope=Envelope)
def send(c, envelope, dest):
    envelope_json = envelope.to_json()
#    print('send: %s' % envelope_json)
    limit, chunk_size = 300, 150
    try:
        if len(envelope_json) < limit:
            raw = [envelope_json]
        else:
            raw = fragmentize(envelope_json, max_chunk_length=chunk_size)

        for r in raw:
            if len(r) > 415:
                print('Do not expect a string of len %s to be ok.' % len(r))
            c.send_message(dest, r)
    except MessageTooLong as e:
        print(e)
        msg = 'Envelope size is %s' % len(envelope_json)
        print(msg)



from Queue import Empty
import argparse
from collections import OrderedDict
from duckietown_swarm import __version__
from duckietown_swarm.utils import get_tuple
import getpass
import json
import random
import subprocess
import sys

from contracts import contract
from contracts.utils import raise_wrapped
from system_cmd import CmdException
from termcolor import colored
import yaml

from .brain import DistributedRepoSwarm
from .constants import DSwarmConstants
from .dcache import Envelope, interpret_message
from .dcache_wire import create_command_message
from .dcache_wire import create_request_message, MsgPong
from .ipfs_utils import IPFSInterface
from .irc2 import duckietown_swarm_main
from .packaging import find_more_information
from .ping import duckietown_swarm_ping_main
from .socket_utils import AsyncLineBasedConnection
from .utils import duration_compact
from .utils import now_until
from duckietown_swarm.packaging import UnexpectedFormat


@contract(returns=DistributedRepoSwarm, patterns=list)
def get_snapshot(host, port, patterns):

    albc = AsyncLineBasedConnection(host, port)
    maddr_from = '/cli'

    inbound = albc.get_q_inbound()
    outbound = albc.get_q_outbound()

    msg = create_request_message(patterns, validity=now_until(105))

    e = Envelope(maddr_from=maddr_from,
                 maddr_to='/dswarm/*/brain',
                 contents=msg)

    outbound.put(e)

    dc = DistributedRepoSwarm()
    from_channel = 'stdin'
    i = 0

#    wait_since_last = 15.0
#    t_last = time.time()
    while True:
        try:
            env = inbound.get(block=True, timeout=1.0)
#            t_last = time.time()
            is_signature = 'signature' in env.contents
            if is_signature:
                sys.stderr.write('S')
            else:
                sys.stderr.write('r')
        except Empty:
            pass
#            if i > 0:
#                if time.time() - t_last > wait_since_last:
#                    break
        else:
#            print env
            msg = interpret_message(env.contents)
            if isinstance(msg, MsgPong):
                break
            dc.process(env.contents, from_channel)
        i += 1
    print('processed %s messages' % i)
    return dc


commands = OrderedDict()


def command(x):
    commands[x.__name__.replace('cmd_', '')] = x
    return x


@command
def cmd_version(parsed, args):  #@UnusedVariable
    print('%s' % __version__)


@command
def cmd_network(parsed, args):  #@UnusedVariable

    host = parsed.host
    port = DSwarmConstants.port_dswarm

    G = get_network(host, port)
    t = get_visjs(G)
    fn = 'network.html'
    with open(fn, 'w') as f:
        f.write(t)
    print('written %s' % fn)


def get_visjs(G):

    edges_all = []
    edges_direct = []
    nodes = []

    t = []

    def id_from_mh(x):
        if not x in t:
            t.append(x)
        return t.index(x)

#    print list(G.nodes_iter(data=True))
    for mh, data in G.nodes_iter(data=True):
        label = data.get('label', mh[-6:])
        nodes.append(dict(id=id_from_mh(mh), label=label))
        for n in G.neighbors(mh):
            direct = G.get_edge_data(mh, n)['direct']
            edges_all.append({'from': id_from_mh(mh), 'to': id_from_mh(n), 'dashes': not direct})
            if direct:
                edges_direct.append({'from': id_from_mh(mh), 'to': id_from_mh(n), 'dashes': not direct})

    t = template

    t = t.replace('EDGES_ALL', json.dumps(edges_all))
    t = t.replace('EDGES_DIRECT', json.dumps(edges_direct))
    t = t.replace('NODES', json.dumps(nodes))
    return t


#    gv.newLink(mh, mh2)
#    gv.newItem(mh)
template = '''
<!doctype html>
<html>
<head>
  <title>Network | Basic usage</title>

  <script type="text/javascript" src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.js"></script>
  <link href="../../dist/vis-network.min.css" rel="stylesheet" type="text/css" />

  <style type="text/css">
    .networks {
      width: 800px;
      height: 800px;
      border: 1px solid lightgray;
    }
  </style>
</head>
<body>

<h3>All connections</h3>

<div class='networks' id="graph_all"></div>

<h3>Only direct connections</h3>
<div  class='networks' id="graph_direct"></div>

<script type="text/javascript">
  // create an array with nodes
  var nodes = new vis.DataSet(NODES);

  // create an array with edges
  var edges_all = new vis.DataSet(EDGES_ALL);
  var edges_direct = new vis.DataSet(EDGES_DIRECT);

  // create a network
  var container = document.getElementById('graph_all');
  var options = {};
  var network1 = new vis.Network(container, {nodes: nodes, edges:edges_all},options);
  var container = document.getElementById('graph_direct');
  var network2 = new vis.Network(container, {nodes: nodes, edges:edges_direct},options);

</script>


</body>
</html>'''


def get_network(host, port):
    what = ['*ipfs_peers', '/pri/*/info/hostname']
    dc = get_snapshot(host, port, what)
    return get_network_(dc)


def get_network_(dc):
    data = dc.summary_dict()
    print(yaml.dump(data))
    import networkx as nx
    G = nx.DiGraph()
    pri = data.get('pri', {})
    hosts = list(pri)
    mh2hostname = {}
    for mh in hosts:
        if 'info' in pri[mh] and 'hostname' in pri[mh]['info']:
            mh2hostname[mh] = pri[mh]['info']['hostname'][0]
        else:
            mh2hostname[mh] = mh[-6:]

        peers = [get_peer(_) for _ in pri[mh].get('ipfs_peers', [])]
#        print '%s->%s' % (mh, peers)

        G.add_node(mh, label=mh2hostname[mh])
        for direct, mh2 in peers:
            G.add_edge(mh, mh2, direct=direct)
    return G


def get_peer(con):
#    /ipfs/QmVKMTGqtgDHMSrGDQnpodnQq4C84kH2Y9v9FGwN8XPtxa/p2p-circuit/ipfs/QmYZQjusVBHq7uWWWrpcghCF1Nyfa3daD3EyfLNdcmERCR,
#      /ipfs/QmVKMTGqtgDHMSrGDQnpodnQq4C84kH2Y9v9FGwN8XPtxa/p2p-circuit/ipfs/QmdcWArNAMwW1veKYvCaSuVsm6j6zd3UWCMpH6i8ZnPuY9
    x = con.split('/')
    p = x[-1]
    assert p.startswith('Qm')
    direct = con.startswith('/ip4')
    return direct, p


@command
def cmd_summary(parsed, args):
    ''' Writes summary'''
    host = parsed.host
    port = DSwarmConstants.port_dswarm

    if args:
        what = args
    else:
        what = ['*']

    dc = get_snapshot(host, port, what)
    data = dc.summary_dict()

    print(yaml.dump(data))


@command
def cmd_online(parsed, args):  #@UnusedVariable
    host = parsed.host
    port = DSwarmConstants.port_dswarm

    what = ['/pri/*/version/dswarm', '/pri/*/info/hostname']

    dc = get_snapshot(host, port, what)
    data = dc.summary_dict(True, True)
    print
    h = []
    hosts = list(data.get('pri', {}))
    for mh in hosts:
        info = data['pri'][mh]['info']
        if 'hostname' in info:
            hostname = info['hostname']['value']
        else:
            hostname = '?'
        if 'version' in info and 'dswarm' in info['version']:
            version = info['version']['dswarm']['value']
            age = info['version']['dswarm']['age']
        else:
            version = '?'
            age = 0

        h.append(dict(version=version, age=age, hostname=hostname, mh=mh))

    h = sorted(h, key=lambda x: x['age'])
    for x in h:
        print('%s   %7s  %10s   %s' % (x['mh'], x['version'], duration_compact(x['age']), x['hostname'],))

#    print(yaml.dump(data))


@command
def cmd_help(parsed, args):  #@UnusedVariable
    print("Commands available:\n")
    for name, cmd in commands.items():
        use = 'dt-swarm %s ' % colored(name, attrs=['bold'])
        doc = cmd.__doc__
        if doc is not None: doc = doc.strip()
        else:
            doc = ''
        print('   %-20s %s' % (use, doc))


@command
def cmd_report(parsed, args):
    host = parsed.host
    port = DSwarmConstants.port_dswarm

    what = ['/%s/*' % DSwarmConstants.SHARE]
    dc = get_snapshot(host, port, what)
    ipfsi = IPFSInterface(None)

    if args:
        find_provs = True
        provs_timeout = args.pop(0)
    else:
        find_provs = False
        provs_timeout = None

    s = create_safe_index(ipfsi, dc, find_provs, provs_timeout)
    mh = ipfsi.add(s)
    url = 'http://localhost:6080/ipfs/' + mh

    print url


@command
def cmd_ipfs(_parsed, args):
    ipfsi = IPFSInterface(None)
    ipfs = ipfsi.get_executable()
    cmd = [ipfs] + args
    p = subprocess.Popen(
            cmd,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            bufsize=0,
            cwd='.',
            env=ipfsi._get_env())
    p.wait()


@command
def cmd_ping(_, _a):
    duckietown_swarm_ping_main()


@command
def cmd_daemon(_parsed, _args):
    duckietown_swarm_main()


def duckietown_swarm_cli_main():

    if len(sys.argv) == 1:
        cmd_help(None, [])
        return

    parser = argparse.ArgumentParser(description='Ping the swarm')
    parser.add_argument('--host', type=str, help='host to use', default='localhost')
    parser.add_argument('command', type=str, help='command')
    parser.add_argument('rest', nargs=argparse.REMAINDER, help='rest')
    parsed = parser.parse_args()

    cmd = parsed.command
    args = parsed.rest

    if not cmd in commands:
        cmd_help(parsed, args)
        return

    f = commands[cmd]
    f(parsed, args)


@command
def cmd_quit(parsed, _args):
    parameters = {}
    host = parsed.host
    send_cmd(host, 'quit', parameters)


@command
def cmd_shell(parsed, args):
    host = parsed.host
    ipfsi = IPFSInterface(None)
    hostname = args.pop(0)
    port = DSwarmConstants.port_dswarm

    try:
        peer = resolve_name(host, port, hostname)
    except NoNameFound as e:
        print e
        return

    print('connecting to peer %s' % peer)
    try:
        ipfsi.swarm_connect("/p2p-circuit/ipfs/%s" % peer)
    except CmdException as e:
        print e
    port = random.randint(22000, 23232)

    cmd = ['ipfs', 'p2p', 'stream', 'dial', '--timeout', '60s',
           peer , 'p2p-shell', '/ip4/127.0.0.1/tcp/%d' % port]
    print(" ".join(cmd))
    res = ipfsi._cmd(cmd)
    print(res.stdout)

    cmd = ['ncat', 'localhost', str(port)]
    print(" ".join(cmd))

    p = subprocess.Popen(
                cmd,
                stdout=sys.stdout,
                stderr=sys.stderr,
                stdin=sys.stdin,
                bufsize=0,
                cwd='.')

    try:
        p.wait()

        ret = p.returncode
        print('Daemon exit: %s' % ret)
        if ret:
            raise Exception(ret)
    finally:
        print('Finished')


class NoNameFound(Exception):
    pass


@command
def cmd_ssh(parsed, args):
    host = parsed.host
    ipfsi = IPFSInterface(None)
    spec = args.pop(0)
    if '@' in spec:
        user = spec[:spec.index('@')]
        hostname = spec[spec.index('@') + 1:]
    else:
        user = getpass.getuser()
        hostname = spec

    port = DSwarmConstants.port_dswarm
    try:
        peer = resolve_name(host, port, hostname)
    except NoNameFound:
        return

    print('Connecting to %r at %r' % (user, hostname))

    print('connecting to peer %s' % peer)
    try:
        ipfsi.swarm_connect("/p2p-circuit/ipfs/%s" % peer)
    except CmdException as e:
        print e
    port = random.randint(22000, 23232)

    cmd = ['ipfs', 'p2p', 'stream', 'dial', '--timeout', '60s',
           peer , 'p2p-ssh', '/ip4/127.0.0.1/tcp/%d' % port]
    print(" ".join(cmd))
    res = ipfsi._cmd(cmd)
    print(res.stdout)

    cmd = ['ssh', '-p', str(port), '-o', 'StrictHostKeyChecking=no',
           '%s@localhost' % user]
    print(" ".join(cmd))

    p = subprocess.Popen(
                cmd,
                stdout=sys.stdout,
                stderr=sys.stderr,
                stdin=sys.stdin,
                bufsize=0,
                cwd='.')

    try:
        p.wait()

        ret = p.returncode
        print('Daemon exit: %s' % ret)
        if ret:
            raise Exception(ret)
    finally:
        print('Finished')


def resolve_name(host, port, hostname):
    what = ["/pri/*/info/hostname"]
    dc = get_snapshot(host, port, what)
    data = dc.summary_dict()
    hostname2peerid = {}
    for peer, values in data['pri'].items():
        _hostname = list(values['info']['hostname'])[0]
        hostname2peerid[_hostname] = peer

    if not hostname in hostname2peerid:
        msg = ('No host %r found in %s' % (hostname, hostname2peerid))
        raise NoNameFound(msg)

    return hostname2peerid[hostname]


def send_cmd(host, name, parameters):
    port = DSwarmConstants.port_dswarm
    albc = AsyncLineBasedConnection(host, port)
    maddr_from = '/cli'

    inbound = albc.get_q_inbound()
    outbound = albc.get_q_outbound()

    msg = create_command_message(name, parameters, validity=now_until(105))

    e = Envelope(maddr_from=maddr_from,
                 maddr_to='/dswarm/*/brain',
                 contents=msg)

    outbound.put(e)

    reply = inbound.get()
    print reply


def create_safe_index(ipfsi, dc, find_provs, provs_timeout):
    ''' Creates a webpage showing the safe data uploaded '''
    data = dc.summary_dict()

    print(yaml.dump(data))

    s = "<html><head></head><body>"

    verified = []

    data_files = get_tuple(data, DSwarmConstants.DLC_BUCKET_FILES, set())
    data_safe = get_tuple(data, DSwarmConstants.DLC_BUCKET_SAFE, set())
    data_verified = get_tuple(data, DSwarmConstants.DLC_BUCKET_VERIFIED, set())
    data_discard = get_tuple(data, DSwarmConstants.DLC_BUCKET_DISCARD, set())

    all_mh = set(data_files) | set(data_safe) | set(data_verified) | set(data_discard)

    for mh in all_mh:
#        s += '\n<a href="/ipfs/%s">%s</a>' % (v, v)
        try:
            more_info = find_more_information(ipfsi, mh, find_provs=find_provs,
                                              provs_timeout=provs_timeout)
            if find_provs:
                print('%15s provided by %s' % (more_info.filename, more_info.providers_payload))
            sys.stderr.write(',')
        except UnexpectedFormat as e:
            print(e)
            continue
        except Exception as e:
            msg = 'Could not get more information about %s' % mh
            raise_wrapped(Exception, e, msg)
#            print(e)
        else:
            verified.append(more_info)

#            if len(verified) > 10:
#                break
    verified = sorted(verified, key=lambda mi: mi.ctime)

    s += '''
        <style>
        .mh { font-family: monospace; }
        thead { font-weight: bold; }
        td { padding-left: 1em; }
        </style>

        <table>
        <thead><tr><td>Filename</td><td>Size</td><td>Upload by</td><td>node</td>
        <td></td>
        <td>proposed</td>
        <td>verified</td>
        <td>safe</td>
        <td>discard</td>


        <td>ipfs_container</td>
        <td>ipfs_payload</td>
        <td>providers</td>

        </tr></thead>
        <tbody>
    '''
    for more_info in verified:
        d = more_info._asdict()
        d['is_proposed'] = 'proposed' if more_info.ipfs in data_files else ''
        d['is_safe'] = 'safe' if more_info.ipfs in data_safe else ''
        d['is_verified'] = 'verified' if more_info.ipfs in data_verified else ''
        d['is_discarded'] = 'discard' if more_info.ipfs in data_discard else ''
        d['date'] = more_info.ctime.strftime('%Y-%m-%d %H:%M')

        def shorten(mh):
            return '<a class="mh" href="/ipfs/%s">%s</a>' % (mh, mh[:4] + '...' + mh[-4:])

        def shorten_node(mh):
            return '<a class="mh" href="/ipns/%s">%s</a>' % (mh, mh[:4] + '...' + mh[-4:])

        d['upload_node_short'] = shorten_node(more_info.upload_node)
        d['ipfs_short'] = shorten(d['ipfs'])
        d['ipfs_payload_short'] = shorten(d['ipfs_payload'])

        d['providers'] = " ".join(map(shorten_node, more_info.providers_info))
        si = '''\n<tr><td><a href="/ipfs/{ipfs_payload}">{filename}</td><td>{size}</td>
        <td>{date}</td>
             <td></td><td> {upload_node_short}</td>
             <td>{is_proposed}</td>
             <td>{is_verified}</td>
             <td>{is_safe}</td>
             <td>{is_discarded}</td>
              <td>{ipfs_short}</td>
              <td>{ipfs_payload_short}</td>
                <td>{providers}</td>
              </tr> '''

        s += si.format(**d)
#        ipfs='QmTJ6sFnB2Pxe1SG7hkd84gkv1YcaEai2tP86mbJKAdBdz', ipfs_payload='QmZYND3nJUW7KAmFMKNjHRMPmtL3FZtqSXW9qNRNcownoG', ipfs_info='QmbWEnQoAL8hPsk2GftZAYC4kTjK7HFYp6eM4D7ULyKA5r', filename='stasera.txt', upload_host='dorothy-7.local', upload_node='QmQDCfGCaYk7Uqc27k6gfgYD13pjny6MZSFQDk8Ra651Jo', upload_user='andrea', ctime=datetime.datetime(2018, 2, 7, 20, 49, 41), size=26, providers_payload=[], providers_info=[])

    s += '</tbody></table>'
    s += '</body>'
    s += '</html>'
    return s


if __name__ == '__main__':
    duckietown_swarm_cli_main()



import time

from dateutil.parser import parse


class DSwarmConstants(object):
    port_dswarm = 8000
#    port_ssh = 8001
    port_shell = 8002
    DELTA_PEER = 10 * 60
    SDELTA_MAKE_SUMMARY = '5m'

    SDELTA_INFO = '5d'

    ALL_BRAINS = '/dswarm/*/brain'

    IPFS_ADDRESSES = 'ipfs_addresses'
    IPFS_PEERS = 'ipfs_peers'

    DEFAULT_DIR = '~/shared-logs'

    PRI = 'pri'

    # IRC_SERVER = 'irc.freenode.net'
    IRC_SERVER = 'frankfurt.co-design.science'
    IRC_CHANNEL = '#duckiebots'

    PUBSUB_TOPIC = 'duckiebots'

    PUBSUB_FRIENDS_GREETING = 'duckiebots\n'

    SHARE = 'share'
    DLC_BUCKET_PEER = 'peer'
    DLC_BUCKET_NETWORKS = 'networks'

    DLC_BUCKET_FILES = (SHARE, 'files')
    DLC_BUCKET_VERIFIED = (SHARE, 'verified')
    DLC_BUCKET_SAFE = (SHARE, 'safe')
    DLC_BUCKET_DISCARD = (SHARE, 'discard')

    BOOTSTRAP_PATTERNS = ['/peers', '*/summaries', '*/ipfs_addresses',
#                          '*/info/*', '*/versions/*'
                          ]

    port_ipfs_api = 6001
    port_ipfs_gw = 6080
    port_ipfs_swarm = 6002

    use_irc = True
    use_pubsub = True
    use_udp = True
    activate_shell_access = False

    date = "Sat Mar  3 10:17:50 CET 2018"
    ignore_before = time.mktime(parse(date).timetuple())




from duckietown_swarm.constants import DSwarmConstants
import json
import os
import platform

from contracts.utils import indent
from system_cmd import system_cmd_result

import subprocess32 as subprocess

from .process_manager import ProcessManager
from .utils import download_url_to_file, get_sha256_for_file_hex


def check_ipfs_executable(repo_dir):
    ipfs = get_ipfs_executable(repo_dir)
    cmd = [ipfs, '--version']
    res = system_cmd_result(cwd='.', cmd=cmd, raise_on_error=True)
    version = res.stdout
    version = version.replace('ipfs', '')
    version = version.replace('version', '')
    version = version.strip()
    print('Detected version %r at %s' % (version, ipfs))
    return version


def get_ipfs_executable(dirname):
    files = {
        'ipfs-0.4.14-dev-armv7l-Linux':
        {
            'mh': 'QmbCLdbcdDYo6KeVnwg6EngqPrm7UVXgJhyaGCPiw1et2f',
            'sha256_hex': '5c185d6d876dc63758a045c2bcd8a924c1050d2eb3d20e328b23ecf147a6e9ba',
        },
        'ipfs-0.4.14-dev-x86_64-Linux':
        {
        'sha256_hex': '7c6eda655b01d09ad5dde69afd4d13ba49953c544f1286831cb96688063422db',
        'mh': 'QmNQtAkXfRchLNt3FuCw3p2kULBEbaofY4WT3HTQMmnbJf',
        },
        'ipfs-0.4.14-dev-x86_64-Darwin': {
        'sha256_hex': 'eb711715ade1deed4508bed61ec63e62ea4fe13072d0399ff243f6e01bbe9821',
        'mh': 'QmcGM8LMNAPWEcJtyan6JovQJHBMs5LDo6Lnza7nwNbJjo',
        },
#
#        'ipfs-0.4.13-x86_64-Darwin':
#        {
#        'sha256_hex': 'eb711715ade1deed4508bed61ec63e62ea4fe13072d0399ff243f6e01bbe9821',
#        'mh': 'QmcGM8LMNAPWEcJtyan6JovQJHBMs5LDo6Lnza7nwNbJjo',
#        }

    }
    choices = {
        ('armv7l', 'Linux'): 'ipfs-0.4.14-dev-armv7l-Linux',
        ('x86_64', 'Linux'): 'ipfs-0.4.14-dev-x86_64-Linux',
        ('x86_64', 'Darwin'): 'ipfs-0.4.14-dev-x86_64-Darwin',
    }
    machine = platform.machine()
    system = platform.system()
    key = (machine, system)
    if not key in choices:
        msg = 'Could not find choice for %s in %s' % (key, choices)
        raise Exception(msg)

    filename = choices[key]
    mh = files[filename]['mh']
    sha256_hex = files[filename]['sha256_hex']

    fn = os.path.join(dirname, filename)
    if not os.path.exists(fn):
        print('Could not find %s' % fn)
        url = 'https://gateway.ipfs.io/ipfs/%s' % mh
        download_url_to_file(url, fn)

    os.chmod(fn, 0744)

    sha256_hex_found = get_sha256_for_file_hex(fn)
    if sha256_hex != sha256_hex_found:
        msg = 'Found corrupted file. \nFound hash: %s \ninstead of %s' % (sha256_hex, sha256_hex_found)
        msg += '\n\nDelete corrupted file %s' % fn
        raise Exception(msg)

    return fn


def initialize_ipfs(repo_dir):
    env = os.environ.copy()
    env['IPFS_PATH'] = repo_dir

    check = os.path.join(repo_dir, 'keystore')
    if not os.path.exists(check):
        print('Creating repo using ipfs init')
        ipfs = get_ipfs_executable(repo_dir)
        cmd = [ipfs, 'init']
        d = os.path.dirname(repo_dir)
        if not os.path.exists(d):
            os.makedirs(d)

        system_cmd_result(cwd='.', cmd=cmd, env=env, raise_on_error=True)
    else:

        print('Repo exists: %s' % check)


def _run_ipfs(repo_dir, use_pubsub):

#    Addresses": {
#    "API": "/ip4/127.0.0.1/tcp/5001",
#    "Announce": [],
#    "Gateway": "/ip4/127.0.0.1/tcp/8080",
#    "NoAnnounce": [],
#    "Swarm": [
#      "/ip4/0.0.0.0/tcp/4001",
#      "/ip6/::/tcp/4001"
#    ]
#  },
    def set_config(name, data):
        filename = os.path.join(repo_dir, 'config')
        with open(filename) as f:
#            contents = f.read()
            config = json.load(f)

        names = name.split('.')

        def set_config(c, name, data):
            one, rest = name[0], name[1:]
            if not one in c:
                msg = 'Could not find config %r in %s' % (one, c)
                raise KeyError(msg)
            if not rest:
                c[one] = data
            else:
                set_config(c[one], rest, data)

        set_config(config, names, data)

        with open(filename, 'w') as f:
            f.write(json.dumps(config))

#        cmd = ['ipfs', 'config', name, '--json', json.dumps(data)]
#        env = os.environ.copy()
#        env['IPFS_PATH'] = repo_dir
#        system_cmd_result(cwd='.', cmd=cmd, env=env, raise_on_error=True)
#        print('Set %s = %s' % (name, data))

    set_config('Addresses.API', '/ip4/0.0.0.0/tcp/%s' % DSwarmConstants.port_ipfs_api)
    set_config('Addresses.Gateway', '/ip4/0.0.0.0/tcp/%s' % DSwarmConstants.port_ipfs_gw)
    port_swarm = DSwarmConstants.port_ipfs_swarm
    set_config('Addresses.Swarm', ["/ip4/0.0.0.0/tcp/%s" % port_swarm ])
    set_config('Swarm.EnableRelayHop', True)
    set_config('Experimental.Libp2pStreamMounting', True)
    set_config('Swarm.ConnMgr.LowWater', 10)
    set_config('Swarm.ConnMgr.HighWater', 20)
    bootstrap_addresses = [
        # frankfurt
        '/ip4/35.156.29.30/tcp/%s/ipfs/Qmew3KSPF7oUdkj8mGoBLd4cRFDKAKfGSdFL7ze7tF8XjJ' % port_swarm,
        '/dns4/frankfurt.co-design.science/tcp/%s/ipfs/Qmew3KSPF7oUdkj8mGoBLd4cRFDKAKfGSdFL7ze7tF8XjJ' % port_swarm,
        # singapore
        '/ip4/139.162.15.229/tcp/%s/ipfs/QmVKMTGqtgDHMSrGDQnpodnQq4C84kH2Y9v9FGwN8XPtxa' % port_swarm,
        '/ip4/singapore.co-design.science/tcp/%s/ipfs/QmVKMTGqtgDHMSrGDQnpodnQq4C84kH2Y9v9FGwN8XPtxa' % port_swarm,
        # rudolf
        "/ip4/129.132.0.35/tcp/%s/ipfs/QmdcWArNAMwW1veKYvCaSuVsm6j6zd3UWCMpH6i8ZnPuY9" % port_swarm,
        '/ip4/idsc-rudolf.ethz.ch/tcp/%s/ipfs/QmdcWArNAMwW1veKYvCaSuVsm6j6zd3UWCMpH6i8ZnPuY9' % port_swarm,

    ]
    set_config('Bootstrap', bootstrap_addresses)

    swarm_key = \
'''/key/swarm/psk/1.0.0/
/base16/
377f1ab022c885ab554482599939b559791442c699c24d8dcf086ab294b8f735'''

#
    env = os.environ.copy()
    env['IPFS_PATH'] = repo_dir

    swarm_key_fn = os.path.join(repo_dir, 'swarm.key')
    with open(swarm_key_fn, 'w') as f:
        f.write(swarm_key)
#    cmd = ['ipfs', 'bootstrap', 'add']
#    system_cmd_result(cwd='.', cmd=cmd, env=env)

    logdir = os.path.join(repo_dir, 'logs')
    if not os.path.exists(logdir):
        os.makedirs(logdir)

    stderr_file = os.path.join(logdir, 'stderr.log')
    stdout = open(os.path.join(logdir, 'stdout.log'), 'w')
    stderr = open(stderr_file, 'w')
    print('Starting daemon')
    ipfs = get_ipfs_executable(repo_dir)
    cmd = [ipfs, 'daemon']

    if use_pubsub:
        cmd.append('--enable-pubsub-experiment')

#    cmd.append('--routing=dhtclient')

    p = subprocess.Popen(
                cmd,
                stdout=stdout,
                stderr=stderr,
                bufsize=0,
                cwd='.',
                env=env)

    try:
        p.wait()

        ret = p.returncode
        print('Daemon exit: %s' % ret)
        if ret:
            stderr.close()
            stdout.close()
            out = open(stderr_file).read()
            msg = "Exited with: \n\n " + indent(out, '>  ')
            raise Exception(msg)
    finally:
#        stderr.write('Finished.')
#        stdout.write('Finished.')
        stderr.close()
        stdout.close()
        print('Finished')


def run_ipfs(repo_dir, use_pubsub):
    args = (repo_dir, use_pubsub)
    pm = ProcessManager(_run_ipfs,
                                args, 'run_ipfs',
                                restart=True)
    pm.start()




from duckietown_swarm.constants import DSwarmConstants
from duckietown_swarm.dcache_wire import create_request_message
from duckietown_swarm.utils import now_until
import socket
import time

import netifaces

from .dcache import CouldNotReadEnvelope
from .dcache import Envelope
from .dcache_wire import create_propose_message

UDP_PORT = 1240


def get_suitable_udp_interfaces():
    ifaces = netifaces.interfaces()

    B = {}

    for iface in ifaces:
        iface = str(iface)
        # exclude docker interfaces
        if 'docker' in iface:
            continue
#        print('--- %s --- ' % iface)
        addrs = netifaces.ifaddresses(iface)
        if netifaces.AF_INET in addrs:
            ip4 = addrs[netifaces.AF_INET]
            for _ in ip4:
                if 'broadcast' in _:
                    B[iface] = {'address': str(_['addr']),
                                'broadcast': str(_['broadcast'])}

#        if netifaces.AF_LINK in addrs:
#            MAC = addrs[netifaces.AF_LINK]
#            print MAC

#    print('Broadcast addresses: %s' % B)
    return B


def get_mac_addresses():
    '''
        returns map ifname -> MAC
    '''
    ifaces = netifaces.interfaces()

    B = {}
    for iface in ifaces:
        addrs = netifaces.ifaddresses(iface)
        iface = str(iface)
        if netifaces.AF_LINK in addrs:
            addresses = addrs[netifaces.AF_LINK]
# https://pypi.python.org/pypi/netifaces
# You might be wondering why this value is a list. The reason is that it's
#  possible for an interface to have more than one address, even within the same family.
#  I'll say that again:
# you can have more than one address of the same type associated with each interface.
#
# Asking for "the" address of a particular interface doesn't make sense.

            for i, _ in enumerate(addresses):
                MAC = _['addr']

                if len(addresses) == 1:
                    name = iface
                else:
                    name = iface + ['a', 'b', 'c', 'd'][i % 4]
                if not is_all_zeros(str(MAC)):
                    B[name] = str(MAC)
    return B


def is_all_zeros(s):
    found = set()
    for char in s:
        if char.isdigit():
            found.add(char)
    return found != set([0])


def udp_broadcaster(mi, interface, broadcast_port):
    addr = mi.my_maddr

    interfaces = get_suitable_udp_interfaces()
    if not interface in interfaces:
        msg = 'No interface %s available' % interface
        raise Exception(msg)

    broadcast_address = interfaces[interface]['broadcast']
    my_address = interfaces[interface]['address']
    ipfsi = mi.get_ipfsi()
    identity = ipfsi.ipfs_id()
    bucket = ('pri', identity, 'udp')
    t = time.time()
    msg = create_propose_message(bucket, addr, validity=[t, t + 60 * 10])
    mi.send_to_brain('', msg, sign=True)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    port = UDP_PORT
    while True:
        try:
            sock.bind((my_address, port))  # is add correct?
        except Exception as e:
            print('could not bind to %s:%s' % (my_address, port))
            print e
            port += 1
            if port > 2000:
                raise
            continue

        if port != UDP_PORT:
            print('finally bound to %s:%s' % (my_address, port))

#        print('ok for %s' % port)
        break

    msg = create_request_message(DSwarmConstants.BOOTSTRAP_PATTERNS, validity=now_until(105))
    maddr_from = '/dswarm/%s/brain' % ipfsi.ipfs_id()
    # otherwise I will answer to myself
#    maddr_from = '/dswarm/*/brain'  #% ipfsi.ipfs_id()
    envelope = Envelope(maddr_from=maddr_from,
                 maddr_to='/dswarm/*/brain',
                 contents=msg)
    q = mi.name2queue[mi.my_maddr]
    q.put(envelope)

    while True:
        envelope = mi.get_next_for_me()
        envelope_json = envelope.to_json()
        sock.sendto(envelope_json, (broadcast_address, broadcast_port))


def udp_listener(mi, interface, broadcast_port):

#    m1 = Multiaddr(mi.my_maddr)
#
#    # get the multiaddr protocol description objects
#    p = m1.protocols()

#    # [Protocol(code=4, name='ip4', size=32), Protocol(code=17, name='udp', size=16)]
#    assert p[0].name == 'ip4'
#    assert p[1].name == 'udp'
    # /ip4/IP/udp/port
#
#    broadcast_address = m1.value_for_protocol(p[0].code)
#    broadcast_port = int(m1.value_for_protocol(p[1].code))

    interfaces = get_suitable_udp_interfaces()

    if not interface in interfaces:
        msg = 'No interface %s available' % interface
        raise Exception(msg)

    broadcast_address = interfaces[interface]['broadcast']
#    my_address = interfaces[interface]['my_address']

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    try:
        sock.bind((broadcast_address, broadcast_port))
    except Exception as e:
        msg = 'Could not bind UDP listener to %s:%s' % (broadcast_address, broadcast_port)
        msg += ' ' + str(e)
        raise Exception(msg)

    t = time.time()
    msg = create_propose_message(('pri', mi.key.hash, 'addresses'),
                                 mi.my_maddr, validity=[t, t + 300])
    mi.send_to_brain('', msg, sign=True)

    from duckietown_swarm.irc2 import CouldNotDispatch

    while True:
        msg, (_who_host, _who_port) = sock.recvfrom(4096)
        try:
            env = Envelope.from_json(msg)
        except CouldNotReadEnvelope as e:
            print(e)
            continue
        else:
            try:
                mi.dispatch(env)
            except CouldNotDispatch as e:
                pass  # print('Ignoring dispatch error for %s' % env.maddr_to)



import random

import base58

from .dcache import create_signature_message
from .utils import get_sha256_base58


class CryptoKey(object):

    def __init__(self, private, public):
        self.private = private
        self.public = public
        self.hash = get_sha256_base58(public)

    def sign_hash(self, data_hash):
        ''' Computes hash of data and then sign it '''
        #return '%s signed by %s' % (data, self.hash)
        return get_sha256_base58(data_hash + 'signature')

    def generate_sign_message(self, data):
        data_hash = get_sha256_base58(data)
        signer = self.hash
        signature = base58.b58encode(self.sign_hash(data_hash))
#        signature = self.sign_hash(data)
        msg = create_signature_message(data_hash, signer, signature)
        return msg


def get_key(cache_dir):  #@UnusedVariable
#    if not os.path.exists(cache_dir):
#        os.makedirs(cache_dir)
#    fn = os.path.join(cache_dir, '.key.pickle')
#    if not os.path.exists(fn):
    key = CryptoKey(str(random.randint(0, 10000)), str(random.randint(0, 10000)))
#        with open(fn, 'wb') as f:
#            pickle.dump(key0, f)
#    with open(fn) as f: key = pickle.load(f)

#    identity = key.hash
    return key



import os
import random
import sys
import time

if __name__ == '__main__':
    print('%s <where> <interval>' % sys.argv[0])
    where = sys.argv[1]
    interval = int(sys.argv[2])
    while True:
        basename = 'dummy%d' % random.randint(1, 100000)
        size = 100 * 1000
        fn = os.path.join(where, basename)
        with open(fn, 'wb') as f:
            for i in range(size):
                f.write(str(i))
        print('written %s' % fn)

        time.sleep(interval)



#! /usr/bin/env python

from duckietown_swarm.synthetic_data import synthetic_data_writer
from multiprocessing import Queue
import os
import time

from contracts import contract
from contracts.utils import indent

from .constants import DSwarmConstants
from .crypto import get_key
from .dcache import Envelope
from .directory_watcher_imp import directory_watcher
from .ipfs_utils import IPFSInterface
from .irc_talker import start_irc
from .pinner import pinner
from .pinner import verifier
from .process_manager import ProcessManager
from .pubsub_friends import connecting_to_peer
from .pubsub_friends import pubsub_friendship
from .pubsub_processes import pubsub_reader_process, pubsub_writer_process
from .shell import shell_access
from .spawn_ipfs import check_ipfs_executable
from .spawn_ipfs import run_ipfs, initialize_ipfs
from .summary import publish_ipns
from .tcpendpoint import tcplisten
from .udp_interface import UDP_PORT, get_suitable_udp_interfaces, \
    udp_broadcaster, udp_listener
from .utils import get_all_available
from .utils import get_at_least_one

DSC = DSwarmConstants


class Dispatcher(object):

    def __init__(self, prefix):
        self.prefix = prefix
        self.name2queue = {}
        self.my_maddr = None

    @contract(envelope=Envelope)
    def dispatch(self, envelope, sign=False):
        n = 0
        if not envelope.maddr_to:
            em = 'The address is empty.'
            em += "\n" + envelope.verbose()
            raise Exception(em)
#        print('dispatch %s' % envelope.maddr_to)

        for name, q in self.name2queue.items():
            is_match, rest = prefix_match(address=envelope.maddr_to, x=name)
            if is_match:
#                print('Sending %r to pipe %r with rest %r' % (envelope.maddr_to, name, rest))
                maddr_to = rest
                maddr_from = self.my_maddr + envelope.maddr_from
                e = Envelope(maddr_from, maddr_to, envelope.contents)
                if sign:
#                    if 'signature' in envelope.contents:
#                        em = 'Will not sign a signature: %s' % envelope
#                        raise Exception(em)
                    msg_sign = self.key.generate_sign_message(envelope.contents)
                    e_sign = Envelope(maddr_from, maddr_to, msg_sign)
                    q.put(e_sign)
                q.put(e)
#                print('forward to queue %s orig: %s\nforw: %s' % (name, envelope, e))
                n += 1

        if n == 0:
            msg = 'Could not find channel for %r' % envelope.maddr_to
            msg += '\n\n' + indent("\n".join(sorted(self.name2queue)), '  ')
            msg += '\n\nEnvelope:'
            msg += '\n\n' + indent(envelope.verbose(), '  ')
            raise CouldNotDispatch(msg)

    def get_ipfsi(self):
        return IPFSInterface(self.ipfs_path)

    @contract(returns=Envelope)
    def get_next_for_me(self):
        q = self.name2queue[self.my_maddr]
        return q.get()

    @contract(returns='list($Envelope)')
    def get_many_for_me(self, timeout):
        q = self.name2queue[self.my_maddr]
        return get_at_least_one(q, timeout=timeout)

    @contract(returns='list($Envelope)')
    def get_all_available_for_me(self):
        q = self.name2queue[self.my_maddr]
        return get_all_available(q)


class CouldNotDispatch(Exception):
    pass


def prefix_match(address, x):
    '''

    Examples:

        /irc/host/channel /irc/host/channel
        True, ''
        /irc/host/channel/b /irc/host/channel
        True, '/b'
        /irc/*/channel /irc/host/channel
        True, ''

        /irc/*/channel/b /irc/host/channel
        True, '/b'

    '''
    ta = address.split('/')
    tx = x.split('/')

#    print('ta: %s' % ta)
#    print('tx: %s' % tx)

    def match(A, B):
        assert not '*' in B, (A, B)
        return A == B or ('*' == A)

#    for lm in reversed(range(len(tx) + 1)):
    lm = len(tx)
    if len(ta) < len(tx):
        return False, None
    for i in range(lm):
        a = ta[i]
        x = tx[i]
        if not match(a, x):
#                print('not matched with %s because %s != %s' % (lm, a, x))
            return False, None
            break
    else:
#            print('matched up to %s: %r %r' % (lm, ta[:lm], tx[:lm]))
        remain = ta[lm:]
        if remain:
            rest = '/' + '/'.join(remain)
        else:
            rest = ''
        return True, rest


class MessagingInterface(Dispatcher):

    def __init__(self, key, ipfs_path):
        prefix = '/dswarm' + '/' + key.hash
        Dispatcher.__init__(self, prefix)

        self.key = key
        self.ipfs_path = ipfs_path

        APP = prefix
        self.MADDR_BRAIN = APP + '/brain'
        self.MADDR_PINNER = APP + '/pinner'
        self.MADDR_CONNECTOR = APP + '/connector'
        self.MADDR_VERIFIER = APP + '/verifier'
        self.MADDR_IPNS = APP + '/ipns'
        self.MADDR_DIRWATCHER = APP + '/dirwatcher'
        self.MADDR_PUBSUBFRIENDS = APP + '/pubfriends'
        self.MADDR_IRC = '/dns4/%s/tcp/6667/irc' % DSC.IRC_SERVER
        self.IRC_CHANNEL = DSC.IRC_CHANNEL
        self.MADDR_PUBSUBWRITER = APP + '/pubsubwriter'
        self.MADDR_READ_SUMMARIES = APP + '/read_summaries'
        self.MADDR_TCP_SERVER = APP + '/tcpserver'

        self.pubsub_topic = DSC.PUBSUB_TOPIC

        self.MADDR_PUBSUBWRITER = '/ipfs-pubsub/' + self.pubsub_topic

        self.name2queue = {}
        self.name2queue[self.MADDR_PINNER] = Queue()
        self.name2queue[self.MADDR_BRAIN] = Queue()
        self.name2queue[self.MADDR_VERIFIER] = Queue()
        self.name2queue[self.MADDR_IPNS] = Queue()
        self.name2queue[self.MADDR_IRC] = Queue()
        self.name2queue[self.MADDR_PUBSUBWRITER] = Queue()
        self.name2queue[self.MADDR_READ_SUMMARIES] = Queue()
        self.name2queue[self.MADDR_CONNECTOR] = Queue()
        self.name2queue[self.MADDR_TCP_SERVER] = Queue()

        self.udps = []

        for name, x in get_suitable_udp_interfaces().items():
#            _address = x['address']
            broadcast = x['broadcast']
#            a = '/ip4/%s/udp/%s' % (broadcast, UDP_PORT)
            a = '/udp/%s/%s/%s' % (name, broadcast, UDP_PORT)
            self.name2queue[a] = Queue()
            self.udps.append(a)

    def broadcast_to_all_brains(self, msg, sign):

        for c in self._get_broadcasting_channels():
            new_to = c + DSC.ALL_BRAINS
            envelope = Envelope(maddr_from='', maddr_to=new_to, contents=msg,
                                comment='process: ' + self.my_maddr)
            self.dispatch(envelope, sign=sign)

    def _get_broadcasting_channels(self):
        return [
            self.MADDR_PUBSUBWRITER,
            self.MADDR_IRC + '/' + self.IRC_CHANNEL,
        ] + self.udps

    def copy_as(self, my_maddr):
        c = MessagingInterface(self.key, self.ipfs_path)
        c.name2queue.update(self.name2queue)
        c.my_maddr = my_maddr
        return c

    @contract(from_channel=str, msg=str)
    def send_to_brain(self, from_channel, msg, sign=False):
        e = Envelope(from_channel, self.MADDR_BRAIN, msg)
        self.dispatch(e, sign=sign)

    @contract(mh='multihash')
    def send_to_pinner(self, mh):
        e = Envelope('', self.MADDR_PINNER, mh)
        self.dispatch(e)

    def send_to_connector(self, address):
        e = Envelope('', self.MADDR_CONNECTOR, address)
        self.dispatch(e)

    @contract(mh='multihash')
    def send_to_readsummaries(self, mh):
        e = Envelope('', self.MADDR_READ_SUMMARIES, mh)
        self.dispatch(e)

    @contract(mh=str)
    def send_to_verifier(self, mh):
        e = Envelope('', self.MADDR_VERIFIER, mh)
        self.dispatch(e)

    @contract(key=str, mh=str)
    def send_to_ipns_publisher(self, key, mh):
        assert mh.startswith('Qm'), (key, mh)
        e = Envelope('', self.MADDR_IPNS, (key, mh))
        self.dispatch(e)


def duckietown_swarm_main():
    from . import __version__
    print('dt-swarm %s' % __version__)
#    print '''
#
#Security note: this program opens the SSH port to the swarm.
#
#
#'''
##It is designed to be used in a friendly environment.
##
##DO NOT use on production machines.
#
#    time.sleep(3)
#    print('ready?')
#    time.sleep(1)
#    print('go!')
#    time.sleep(1)

    watch_dir = os.path.expanduser(DSwarmConstants.DEFAULT_DIR)

    if not os.path.exists(watch_dir):
        os.makedirs(watch_dir)

    cache_dir = os.path.join(watch_dir, '.cache')

    ipfs_path = os.path.join(cache_dir, '.ipfs')
    check_ipfs_executable(ipfs_path)
    initialize_ipfs(ipfs_path)
    ipfsi = IPFSInterface(ipfs_path)

    try:
        print ipfsi.stats_bw()
        print('OK - ipfs was running')
    except:
        print('Now starting ipfs')
        run_ipfs(ipfs_path, DSwarmConstants.use_pubsub)
        while True:
            time.sleep(1)
            try:
                print ipfsi.stats_bw()
                break
            except:
                pass

    identity = ipfsi.ipfs_id()

    print('%s Identity: %s' % (ipfs_path, identity))

    ipfsi.p2p_listener_open('p2p-ssh', '/ip4/127.0.0.1/tcp/22')
    ipfsi.p2p_listener_open('p2p-dswarm', '/ip4/127.0.0.1/tcp/%s' % DSwarmConstants.port_dswarm)
    ipfsi.p2p_listener_open('p2p-shell', '/ip4/127.0.0.1/tcp/%s' % DSwarmConstants.port_shell)

    key = get_key(cache_dir)
    key.hash = identity
    print('You can put any log file in the directory\n\n\t%s\n\n'
          'and it will be shared with the swarm.\n\n' % watch_dir)

    '''
        incoming -> [brain] -> outgoing
                               pinner_queue
                    publisher_queue


            pinner_queue -> [pinner] -> pinner

            publisher_queue -> [publisher] -> incoming



        outgoing[0] -> [irc server]
        outgoing[1]
    '''

    mi = MessagingInterface(key, ipfs_path)
    interfaces = get_suitable_udp_interfaces()
    if DSwarmConstants.use_udp:
        for x in mi.udps:
            interface = x.split('/')[2]
            assert interface in interfaces, (interface, x.split('/'))
            pm = ProcessManager(udp_broadcaster,
                                (mi.copy_as(x), interface, UDP_PORT), x,
                                restart=True)
            pm.start()

            pm = ProcessManager(udp_listener,
                                (mi.copy_as(x), interface, UDP_PORT), x,
                                restart=True)
            pm.start()

    pm = ProcessManager(publish_ipns,
                        (mi.copy_as(mi.MADDR_IPNS),), 'publish_ipns',
                        restart=True)
    pm.start()

    if DSwarmConstants.use_pubsub:
        pm = ProcessManager(pubsub_reader_process,
                            (mi.copy_as(mi.MADDR_PUBSUBWRITER), mi.pubsub_topic), 'pubsub_reader',
                            restart=True)
        pm.start()

        pm = ProcessManager(pubsub_writer_process,
                            (mi.copy_as(mi.MADDR_PUBSUBWRITER), mi.pubsub_topic), 'pubsub_writer',
                            restart=True)
        pm.start()

    npinners = 10
    for i in range(npinners):

        pm = ProcessManager(pinner, (mi.copy_as(mi.MADDR_PINNER),), 'pinner%d' % i, restart=True)
        pm.start()

    nconnectors = 10
    for i in range(nconnectors):

        pm = ProcessManager(connecting_to_peer, (mi.copy_as(mi.MADDR_CONNECTOR),), 'connector%d' % i, restart=True)
        pm.start()

    nverifiers = 5
    for i in range(nverifiers):
        pm = ProcessManager(verifier, (mi.copy_as(mi.MADDR_VERIFIER),), 'verifier%d' % i, restart=True)
        pm.start()

    from .brain import brain

    # server = ('frankfurt.co-design.science', 6667)
    server = (DSC.IRC_SERVER, 6667)
    channel = '#duckiebots'
    if DSwarmConstants.use_irc:
        pm = ProcessManager(start_irc,
                            (mi.copy_as(mi.MADDR_IRC), server, channel),
                            'irc1',
                            restart=True)
        pm.start()

#    if False:
#        servers = [('irc.freenode.net', 6667), ]
#        pm = ProcessManager(start_irc,
#                            ('to_irc2', servers, mi), 'irc2',
#                            restart=True)
#        pm.start()

    brainp = ProcessManager(brain,
                        (mi.copy_as(mi.MADDR_BRAIN), cache_dir), 'brain',
                        restart=True)
    brainp.start()

    pm = ProcessManager(directory_watcher,
                        (mi.copy_as(mi.MADDR_DIRWATCHER), watch_dir), 'directory_watcher',
                        restart=True)
    pm.start()

    from .offline import read_nodes_summaries

    for i in range(1):
        pm = ProcessManager(read_nodes_summaries,
                            (mi.copy_as(mi.MADDR_READ_SUMMARIES),), 'read_nodes_summaries%d' % i,
                            restart=True)
        pm.start()

    pm = ProcessManager(pubsub_friendship,
                        (mi.copy_as(mi.MADDR_PUBSUBFRIENDS),), 'pubsub_friendship',
                        restart=True)
    pm.start()

    pm = ProcessManager(tcplisten,
                        (mi.copy_as(mi.MADDR_TCP_SERVER), DSwarmConstants.port_dswarm),
                        'tcplisten%s' % DSwarmConstants.port_dswarm,
                        restart=True)
    pm.start()

    if DSwarmConstants.activate_shell_access:
        pm = ProcessManager(shell_access,
                            (DSwarmConstants.port_shell,), 'shell%s' % DSwarmConstants.port_shell,
                            restart=True)
        pm.start()

    if False:
        pm = ProcessManager(synthetic_data_writer, (watch_dir,), 'synthetic', restart=True)
        pm.start()

#    pm = ProcessManager(publisher_summary,
#                        (mi, '/dswarm/summary'), 'publisher_summary',
#                        restart=True)
#    pm.start()

    brainp.p.join()


if __name__ == "__main__":
    duckietown_swarm_main()



import datetime
import os
import socket
import time


def synthetic_data_writer(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    delta = 60
    length = 1024

    i = 0
    while True:
        hname = socket.gethostname()[:8]
        while True:
            basename = 'dummy_%s_%05d.txt' % (hname, i)
            fn = os.path.join(dirname, basename)
            if os.path.exists(fn):
                i += 1
            else:
                break
        contents = str(datetime.datetime.now())
        contents += '\n' + '?' * length

        with open(fn, 'w') as f:
            f.write(contents)
        print('Created %s' % fn)
        time.sleep(delta)




import duckietown_swarm
from duckietown_swarm.dcache_wire import MsgPong
from duckietown_swarm.utils import now_until, now_until_forever
import fnmatch
import pickle
import socket
import sys
import time

from contracts import check_isinstance
from contracts.utils import indent

from .constants import DSwarmConstants
from .dcache import DistributedRepo, ProcessFailure, interpret_message, Envelope
from .dcache_wire import MsgCommand, MsgPing, create_pong_message, \
    MsgRequest, create_dismiss_message, create_propose_message, create_signature_message
from .ipfs_utils import IPFSInterface
from .irc2 import MessagingInterface
from .udp_interface import get_mac_addresses
from .utils import DoAtInterval, get_sha256_base58
from .utils import as_seconds

DSC = DSwarmConstants


class DistributedRepoSwarm(DistributedRepo):

    def __init__(self):
        DistributedRepo.__init__(self)
        self.init_bucket(DSC.DLC_BUCKET_NETWORKS)
        self.init_bucket(DSC.DLC_BUCKET_PEER)

        self.init_bucket(DSC.SHARE)  # XXX: should be automatically created
        self.init_bucket(DSC.DLC_BUCKET_FILES)
        self.init_bucket(DSC.DLC_BUCKET_VERIFIED)
        self.init_bucket(DSC.DLC_BUCKET_SAFE)
        self.init_bucket(DSC.DLC_BUCKET_DISCARD)

        self.init_bucket(DSC.PRI)


def brain(mi, cache_dir):
    check_isinstance(mi, MessagingInterface)

    ipfsi = IPFSInterface(mi.ipfs_path)

    dc = DistributedRepoSwarm()

    pickle.dumps(dc.root)  # test
    dc.use_cache_dir(cache_dir)

    dc.cleanup()
    dc.sync()

#    already_verified = set(dc.query(DSC.DLC_BUCKET_VERIFIED))
#    already_files = set(DSC.DLC_BUCKET_FILES)
#    print('From cache: %s files' % len(already_files))
#    print('From cache: %s verified' % len(already_verified))

    bucket_files = dc.root.get_bucket(DSC.DLC_BUCKET_FILES, False)
    bucket_verified = dc.root.get_bucket(DSC.DLC_BUCKET_VERIFIED, False)
    bucket_safe = dc.root.get_bucket(DSC.DLC_BUCKET_SAFE, False)
    bucket_discard = dc.root.get_bucket(DSC.DLC_BUCKET_DISCARD, False)

    considered_for_verifier = set()

    def on_new_files(d):
        Q = d['data']
        if Q in considered_for_verifier:
            return
        considered_for_verifier.add(Q)
        if Q in bucket_verified.data2dataitem:
            print('new file %s already in /verified' % Q)
            return
        if Q in bucket_safe.data2dataitem:
            print('new file %s already in /safe' % Q)
            return
        if Q in bucket_discard.data2dataitem:
            print('new file %s already in /discard' % Q)
            return

        print('New file to send to verifier: %s ' % d)
        mi.send_to_verifier(Q)

    def on_new_verified(d):
        Q = d['data']
        if Q in bucket_safe.data2dataitem:
            print('verified file %s already in /safe' % Q)
            return
        if Q in bucket_discard.data2dataitem:
            print('verified file %s already in /discard' % Q)
            return

        mi.send_to_pinner(Q)

    def on_new_peer(d):
        data = d['data']
        print('New peer: %s' % data)
        if data != identity:
            address = '/p2p-circuit/ipfs/' + data
            mi.send_to_connector(address)

    def on_new_summary(d):
        bucket = d['bucket_name']
        data = d['data']
        if bucket[-1] == 'summaries':
            #print('New summary: %s' % d)
            which = bucket[1]
            if which != identity:
                print('Look up from %s %s' % (which, data))
                mi.send_to_readsummaries(data)
        else:
            pass

    def on_new_peer_address(d):
        bucket = d['bucket_name']
        data = d['data']
#        print('on_new_peer_address: %s ' % d)
        if bucket[-1] == DSC.IPFS_ADDRESSES:
            which = bucket[-2]
            print('on_new_peer_address: %s %s' % (which, data))
            if which != identity:
                mi.send_to_connector(data)
        else:
            pass

#            print('Ignoring new in bucket %s = %s' % (str(bucket), data))

    def periodic_maintenance():

        def if_in_A_remove_from_B(A, B):
            for Q in A.query():
                if Q in B.data2dataitem:
                    print('periodic maintainance: removing %s' % Q)
                    del B.data2dataitem[Q]

        # if Q is in /verified, remove it from /files
        if_in_A_remove_from_B(bucket_verified, bucket_files)
        # if Q is in /discard, remove it from /verified and /files
        if_in_A_remove_from_B(bucket_discard, bucket_verified)
        if_in_A_remove_from_B(bucket_discard, bucket_files)
        # if Q is in /safe, remove it from  /verified and /files
        if_in_A_remove_from_B(bucket_safe, bucket_verified)
        if_in_A_remove_from_B(bucket_safe, bucket_files)

    print('initial periodic_maintenance')
    periodic_maintenance()
    # initialization for files
    print('Initialize files')
    for p in dc.query(DSC.DLC_BUCKET_FILES):
        on_new_files(dict(data=p))
    print('Initialize verified')
    for p in dc.query(DSC.DLC_BUCKET_VERIFIED):
        on_new_verified(dict(data=p))

    dc.add_hook_proposed(DSC.DLC_BUCKET_PEER, on_new_peer)
    dc.add_hook_proposed(DSC.DLC_BUCKET_FILES, on_new_files)
    dc.add_hook_proposed(DSC.DLC_BUCKET_VERIFIED, on_new_verified)
    dc.add_hook_proposed_all(on_new_summary)
    dc.add_hook_proposed_all(on_new_peer_address)

    interval_make_summary = DoAtInterval(as_seconds(DSwarmConstants.SDELTA_MAKE_SUMMARY))
    interval_advertise_peer = DoAtInterval(DSwarmConstants.DELTA_PEER * 0.5)
    interval_addresses = DoAtInterval(as_seconds('10m'))
    interval_rebroadcast = DoAtInterval(as_seconds('3m'))
    interval_connect = DoAtInterval(as_seconds('5m'))
    interval_advertises_ipfs_peers = DoAtInterval(as_seconds('1m'))
#    interval_cleanup_sync = DoAtInterval(as_seconds('5m'), as_seconds('5m'))
    interval_cleanup_sync = DoAtInterval(as_seconds('5m'))
    interval_queue_stats = DoAtInterval(as_seconds('10s'))

    key = mi.key
    identity = key.hash

    for peer in dc.query(DSC.DLC_BUCKET_PEER):
        if peer != identity:
            print('trying to connect %s' % peer)
            address = '/p2p-circuit/ipfs/' + peer
            mi.send_to_connector(address)

    pickle.dumps(dc.root)  # test

    try:
        should_quit = False
        while not should_quit:
            if interval_advertises_ipfs_peers.its_time():
                periodic_advertise_ipfs_peers(dc, mi, identity, ipfsi)

            if interval_addresses.its_time():
                periodic_advertise_ipfs_addresses(dc, mi, identity, ipfsi)

            if interval_connect.its_time():
                periodic_try_connect(dc, mi, identity, ipfsi)

            if interval_make_summary.its_time():
                print('making summary...')
                periodic_make_summary(dc, mi, identity, ipfsi)

            if interval_advertise_peer.its_time():
                periodic_advertise_self(dc, mi, identity, ipfsi)

            envelopes = mi.get_many_for_me(timeout=2)

            if interval_queue_stats.its_time():
#                msg = 'Processing %d envelopes' % len(envelopes)
#                print(msg)
                pass

            if not envelopes:
                sys.stderr.write('-')

            for envelope in envelopes:
                if identity in envelope.maddr_from:
                    sys.stderr.write('u')  # us
                else:
                    sys.stderr.write('+')  # others

                try:
                    m = interpret_message(envelope.contents)
                except Exception as e:
                    msg = 'Could not interpret message from %s' % envelope.maddr_from
                    msg += '\n\n' + indent(str(e), '> ')
                    print(msg)
                else:
                    if isinstance(m, MsgPing):
                        reply_ping(mi, dc, envelope)
                    elif isinstance(m, MsgPong):
                        pass
                    elif isinstance(m, MsgRequest):
                        reply_msg_request(mi, dc, envelope, m)

                    elif isinstance(m, MsgCommand):
                        if m.command == 'quit':
                            should_quit = True
                            pong = create_pong_message()
                            e = Envelope('', envelope.maddr_from, pong)
                            mi.dispatch(e)
                            time.sleep(5)
                        else:
                            print('cannot interpret %r' % m.command)
                            pong = create_pong_message()
                            e = Envelope('', envelope.maddr_from, pong)
                            mi.dispatch(e)
                    else:
                        try:
                            dc.process(envelope.contents, envelope.maddr_from)
                        except ProcessFailure as e:
                            em = 'Cannot interpret message: '
                            em += '\n' + str(e)
                            print(em)

            if interval_rebroadcast.its_time():
                periodic_rebroadcast(mi, dc)

            if interval_cleanup_sync.its_time():
                periodic_maintenance()
                dc.cleanup()
                dc.sync()

        print('Graceful exit.')
    finally:

        print('Clean up cache')
        dc.close()


last_summary = None


def periodic_advertise_ipfs_addresses(dc, mi, identity, ipfsi):  #@UnusedVariable
    addresses = ipfsi.get_addresses()
    print('addresses: %s' % addresses)
    for address in addresses:
        if (address.startswith('/ip4/127.0.0.1/') or
            address.startswith('/ip6/')):
            continue
        msg = create_propose_message((DSC.PRI, identity, DSC.IPFS_ADDRESSES),
                                     address, validity=now_until(as_seconds('1h')))
        mi.send_to_brain('', msg, sign=True)


def periodic_advertise_ipfs_peers(dc, mi, identity, ipfsi):  #@UnusedVariable
    peers = ipfsi.swarm_peers()
    for peer in peers:
        msg = create_propose_message((DSC.PRI, identity, DSC.IPFS_PEERS),
                                     peer, validity=now_until(as_seconds('1h')))
        mi.send_to_brain('', msg, sign=True)


def periodic_try_connect(dc, mi, identity, ipfsi):  #@UnusedVariable
    for t, bucket in dc.root.items():
        if t[-1] == DSC.IPFS_ADDRESSES:
            if t[-2] == identity: continue
            for address in bucket.query():
                print('connecting %s' % address)
                mi.send_to_connector(address)


def periodic_advertise_self(dc, mi, identity, ipfsi):  #@UnusedVariable
    msg = create_propose_message(DSC.DLC_BUCKET_PEER, identity, validity=now_until(DSwarmConstants.DELTA_PEER))
    mi.send_to_brain('', msg, sign=True)

    import platform

    properties = {}
    properties['hostname'] = socket.gethostname()
    properties['processor'] = platform.processor()
    properties['machine'] = platform.machine()
    # properties['node'] = platform.node() # really same as machines
    properties['version/dswarm'] = duckietown_swarm.__version__
    properties['version/python'] = platform.python_version()
    properties['version/ipfs'] = ipfsi.version()

    interfaces = []
    for iname, mac in get_mac_addresses().items():
        i = '%s;%s' % (iname, mac)
        interfaces.append(i)
    properties['interfaces'] = interfaces
    properties['system'] = platform.system()

    for k, v in properties.items():
        if not isinstance(v, list):
            v = [v]
        for x in v:
            bucket = (DSC.PRI, identity, 'info') + tuple(k.split('/'))

            msg = create_propose_message(bucket, x, validity=now_until(as_seconds(DSC.SDELTA_INFO)))
            mi.send_to_brain('', msg, sign=True)


def periodic_make_summary(dc, mi, identity, ipfsi):
    summary = ipfsi.get_tree_builder()

    global last_summary
    print('dc.summary')
    summary_humans = dc.summary()
    summary_humans = summary_humans.replace(identity, 'QmSELF')
    summary.add_file_content('humans.txt', summary_humans)

    print('dc.summary_messages')
    summary_messages = dc.summary_messages()
    summary.add_file_content('machines.txt', '\n'.join(summary_messages))

    print('network')
    from duckietown_swarm.cli import get_network_, get_visjs
    G = get_network_(dc)
    try:
        network_visualization = get_visjs(G)
        summary.add_file_content('network.html', network_visualization)
    except Exception as e:
        print e

    Q = summary.get_hash()

    bucket = (DSC.PRI, identity, 'summaries')
    if last_summary is not None:
        msg = create_dismiss_message(bucket, last_summary, validity=now_until_forever())
        mi.broadcast_to_all_brains(msg, sign=True)

    validity = now_until(2 * as_seconds(DSwarmConstants.SDELTA_MAKE_SUMMARY))
    msg = create_propose_message(bucket, Q, validity)
    mi.broadcast_to_all_brains(msg, sign=True)

    last_summary = Q
    print('summary: %s' % last_summary)


def periodic_rebroadcast(mi, dc):
    messages = []

    for t, _ in dc.root.items():
        out_channels = mi._get_broadcasting_channels()
        these = dc.rebroadcast(t, out_channels)
        # TODO: make sure that there is a signature
        if t[0] == DSC.PRI:
            for m in these:
                hashed = get_sha256_base58(m.contents)
                if not hashed in dc.signatures:
                    print('No signature available for %s' % "/".join(t))
                else:
                    messages.append(m)

        else:
            messages.extend(these)

    for envelope in messages:
        # find signatures
        hashed = get_sha256_base58(envelope.contents)
        if hashed in dc.signatures:
            for signer, signature in dc.signatures[hashed].items():
                msg = create_signature_message(hashed, signer, signature)
                e_sign = Envelope(envelope.maddr_from, envelope.maddr_to, msg)
                mi.dispatch(e_sign)
        mi.dispatch(envelope)


def reply_msg_request(mi, dc, envelope, m):
    messages = []
    patterns = m.patterns

    if mi.my_maddr in envelope.maddr_from:
        msg = 'Ignoring request from ourselves.'
        # print(msg)
        return
    print('received request: %s' % envelope)
    print('patterns: %s' % str(patterns))

    def matches(x):
        for pattern in patterns:
            if fnmatch.fnmatch(x, pattern):
                return True
        return False

    for bucket_name, _bucket in dc.root.items():
        c = "/" + "/".join(bucket_name)
        if not matches(c):
            continue
        out = dc._summary_message_for_bucket(bucket_name, only_positive=True)

        messages.extend(out)

    print('Sending %d messages in response to query %s' % (len(messages), patterns))
    for msg in messages:
        e = Envelope('', envelope.maddr_from, msg)
        mi.dispatch(e)

    pong = create_pong_message()
    e = Envelope('', envelope.maddr_from, pong)
    mi.dispatch(e)


def reply_ping(mi, _, envelope):
    pong = create_pong_message()
    e = Envelope('', envelope.maddr_from, pong)
    mi.dispatch(e, sign=True)



from setuptools import find_packages, setup


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version(filename='src/duckietown_swarm/__init__.py')

name = 'duckietown_swarm'

setup(name=name,
      url='http://github.com/duckietown/duckietown_swarm',
      maintainer="Andrea Censi",
      maintainer_email="andrea@duckietown.org",
      description='',
      long_description='',
      #package_data={'':['*.*', '*.mcdp*', '*.js', '*.png', '*.css']},

      # without this, the stuff is included but not installed
      include_package_data=True,
      keywords="Optimization",
      license="GPLv2",
      classifiers=[
        'Development Status :: 4 - Beta',
      ],
      version=version,

      download_url=
        'http://github.com/duckietown/duckietown_swarm/tarball/%s' % version,

      package_dir={'':'src'},
      packages=find_packages('src'),
      install_requires=[
        #'ConfTools>=1.0,<2',
        # '#quickapp',
        # 'reprep',
        'irc',
        'SystemCmd',
        'ConfTools',
        'termcolor',
        'ruamel.yaml',
        'ruamel.ordereddict',
        'base58',
        'netifaces',
        'multiaddr',
        'subprocess32',
        'setproctitle',
        'networkx>=1.10,<2',
      ],
      # This avoids creating the egg file, which is a zip file, which makes our data
      # inaccessible by dir_from_package_name()
      zip_safe=False,
      dependency_links=[
      ],

      tests_require=[
#        'nose>=1.1.2,<2',
#        'comptests',
      ],

      entry_points={

         'console_scripts': [
            'dt-swarm = duckietown_swarm:duckietown_swarm_cli_main',
            'dt-swarm-watcher = duckietown_swarm:duckietown_swarm_watcher_main',
        ]
      }
)



import sys
from system_cmd import system_cmd_result


def duckietown_swarm_watcher_main():
    args = sys.argv[1:]
    cmd = ['dt-swarm'] + args
    cwd = '.'
    while True:
        res = system_cmd_result(cwd, cmd, raise_on_error=False,
                          display_stdout=True, display_stderr=True)
        print('res: %s' % res.ret)


if __name__ == "__main__":
    duckietown_swarm_watcher_main()
